"""
finite size scaling tools
"""
from typing import List, Tuple, Optional

import numpy as np


def data_collapse(
    n: List[int],
    p: List[float],
    obs: List[List[float]],
    pc: float,
    nu: float,
    beta: float = 0,
    obs_type: int = 1,
    fit_type: int = 0,
    dobs: Optional[List[List[float]]] = None,
) -> Tuple[List[float], List[List[float]], List[List[float]], float]:
    xL: List[List[float]] = []  # x=(p-pc)L^(1/\nv)
    yL: List[List[float]] = []  # y=S(p,L)-S(pc,L) or S(p,L)
    pc_list = []
    if not isinstance(p[0], list):
        p = [p for _ in n]  # type: ignore
    for n0 in range(len(n)):
        xL.append([])
        yL.append([])
        for p0 in range(len(p[n0])):  # type: ignore
            xL[n0].append((p[n0][p0] - pc) * (n[n0] ** (1 / nu)))  # type: ignore
            pc_L = pc_linear_interpolation(p[n0], obs[n0], pc)  # type: ignore
            if obs_type == 0:
                yL[n0].append((obs[n0][p0] - pc_L) * n[n0] ** beta)
                # entanglement with only collapse and no crossing
            else:
                yL[n0].append(obs[n0][p0] * n[n0] ** beta)
                # tripartite mutual information
        pc_list.append(pc_L)
    if fit_type == 0:
        xL_all = []
        for i in range(len(xL)):
            xL_all.extend(xL[i])
        yL_ave = []
        loss = []
        for x0 in range(len(xL_all)):
            ybar_list = []
            for n0 in range(len(n)):
                if xL_all[x0] >= xL[n0][0] and xL_all[x0] <= xL[n0][-1]:
                    yinsert = pc_linear_interpolation(xL[n0], yL[n0], xL_all[x0])
                    ybar_list.append(yinsert)

            ybar = np.mean(ybar_list)
            mean_squared = [(ybar_list[i] - ybar) ** 2 for i in range(len(ybar_list))]
            loss.append(np.sum(mean_squared))
            yL_ave.append(ybar)
        loss = np.sum(loss)

        return pc_list, xL, yL, loss  # type: ignore

    # fit_type == 1
    if dobs is None:
        raise ValueError("uncertainty of each y has to be specified in `dobs`")

    datas = []
    for n0 in range(len(n)):
        for i in range(len(xL[n0])):
            datas.append([xL[n0][i], yL[n0][i], dobs[n0][i]])
    datas = sorted(datas, key=lambda x: x[0])
    loss = _quality_objective_v2(datas)  # type: ignore
    return pc_list, xL, yL, loss  # type: ignore


def _quality_objective_v2(datas: List[List[float]]) -> float:
    #  https://journals.aps.org/prb/supplemental/10.1103/PhysRevB.101.060301/Supplement.pdf
    loss = []
    for i in range(len(datas) - 2):
        # i, i+1, i+2
        x, y, d = datas[i + 1]
        x1, y1, d1 = datas[i]
        x2, y2, d2 = datas[i + 2]
        if np.abs(x - x1) < 1e-4 or np.abs(x - x2) < 1e-4:
            continue
        ybar = ((x2 - x) * y1 - (x1 - x) * y2) / (x2 - x1)
        delta = (
            d**2
            + d1**2 * (x2 - x) ** 2 / (x2 - x1) ** 2
            + d2**2 * (x1 - x) ** 2 / (x2 - x1) ** 2
        )
        w = (y - ybar) ** 2 / delta
        loss.append(w)
    return np.mean(loss)  # type: ignore


def pc_linear_interpolation(p: List[float], SA: List[float], pc_input: float) -> float:
    if pc_input in p:
        pc_index = p.index(pc_input)
        pc = SA[pc_index]
    else:
        pr_index = 0
        for p0 in range(len(p)):
            if p[p0] > pc_input:
                pr_index = p0
                break

        x = [p[pr_index - 1], p[pr_index]]
        y = [SA[pr_index - 1], SA[pr_index]]
        linear_model = np.polyfit(x, y, 1)
        linear_model_fn = np.poly1d(linear_model)
        pc = linear_model_fn(pc_input)
    return pc



"""
Cloud provider from local machine
"""

from typing import Any, Dict, List, Optional, Union, Sequence
from uuid import uuid4
import time

from .abstraction import Device, sep, Task
from ..utils import is_sequence
from ..abstractcircuit import AbstractCircuit

local_devices = ["testing"]

task_list: Dict[str, Any] = {}  # memory only task cache


def list_devices(token: Optional[str] = None, **kws: Any) -> List[Device]:
    rs = []
    for d in local_devices:
        rs.append(Device.from_name("local" + sep + d))
    return rs


def get_task_details(
    task: Task, device: Device, token: str, prettify: bool
) -> Dict[str, Any]:
    if task.id_ in task_list:
        return task_list[task.id_]  # type: ignore
    raise ValueError("no task with id: %s" % task.id_)


def submit_task(
    device: Device,
    token: str,
    shots: Union[int, Sequence[int]] = 1024,
    version: str = "1",
    circuit: Optional[Union[AbstractCircuit, Sequence[AbstractCircuit]]] = None,
    **kws: Any
) -> List[Task]:
    def _circuit2result(c: AbstractCircuit) -> Dict[str, Any]:
        if device.name in ["testing", "default"]:
            count = c.sample(batch=shots, allow_state=True, format="count_dict_bin")  # type: ignore
        else:
            raise ValueError("Unsupported device from local provider: %s" % device.name)
        d = {
            "id": str(uuid4()),
            "state": "completed",
            "at": time.time() * 1e6,
            "shots": shots,
            "device": device.name,
            "results": count,
        }
        return d

    if is_sequence(circuit):
        tl = []
        for c in circuit:  # type: ignore
            d = _circuit2result(c)
            task_list[d["id"]] = d
            tl.append(Task(id_=d["id"], device=device))
        return tl
    else:
        d = _circuit2result(circuit)  # type: ignore
        task_list[d["id"]] = d

        return Task(id_=d["id"], device=device)  # type: ignore


def list_tasks(device: Device, token: str, **filter_kws: Any) -> List[Task]:
    r = []
    for t, v in task_list.items():
        if (device is not None and v["device"] == device.name) or device is None:
            r.append(Task(id_=t, device=Device.from_name("local" + sep + v["device"])))
    return r



"""
compiler interface via qiskit
"""

from typing import Any, Dict, Optional
import re

from ..abstractcircuit import AbstractCircuit
from ..circuit import Circuit
from ..translation import qiskit_from_qasm_str_ordered_measure


def _free_pi(s: str) -> str:
    # dirty trick to get rid of pi in openqasm from qiskit
    rs = []
    pistr = "3.141592653589793"
    s = s.replace("pi", pistr)
    for r in s.split("\n"):
        inc = re.search(r"\(.*\)", r)
        if inc is None:
            rs.append(r)
        else:
            v = r[inc.start() : inc.end()]
            v = eval(v)
            if not isinstance(v, tuple):
                r = r[: inc.start()] + "(" + str(v) + ")" + r[inc.end() :]
            else:  # u gate case
                r = r[: inc.start()] + str(v) + r[inc.end() :]
            rs.append(r)
    return "\n".join(rs)


def _comment_qasm(s: str) -> str:
    """
    return the qasm str in comment format

    :param s: _description_
    :type s: str
    :return: _description_
    :rtype: str
    """
    nslist = []
    nslist.append("//circuit begins")
    for line in s.split("\n"):
        nslist.append("//" + line)
    nslist.append("//circuit ends")
    return "\n".join(nslist)


def _comment_dict(d: Dict[int, int], name: str = "logical_physical_mapping") -> str:
    """
    save a dict in commented qasm

    :param d: _description_
    :type d: Dict[int, int]
    :param name: _description_, defaults to "logical_physical_mapping"
    :type name: str, optional
    :return: _description_
    :rtype: str
    """
    nslist = []
    nslist.append("//%s begins" % name)
    for k, v in d.items():
        nslist.append("// " + str(k) + " : " + str(v))
    nslist.append("//%s ends" % name)
    return "\n".join(nslist)


def _get_positional_logical_mapping_from_qiskit(qc: Any) -> Dict[int, int]:
    i = 0
    positional_logical_mapping = {}
    for inst in qc.data:
        if inst[0].name == "measure":
            positional_logical_mapping[i] = inst[1][0].index
            i += 1

    return positional_logical_mapping


def _get_logical_physical_mapping_from_qiskit(
    qc_after: Any, qc_before: Any = None
) -> Dict[int, int]:
    """
    get ``logical_physical_mapping`` from qiskit Circuit by comparing the circuit after and before compiling

    :param qc_after: qiskit ``QuantumCircuit`` after compiling
    :type qc_after: Any
    :param qc_before: qiskit ``QuantumCircuit`` before compiling,
        if None, measure(q, q) is assumed
    :type qc_before: Any
    :return: logical_physical_mapping
    :rtype: Dict[int, int]
    """
    logical_physical_mapping = {}
    for inst in qc_after.data:
        if inst[0].name == "measure":
            if qc_before is None:
                logical_q = inst[2][0].index
            else:
                for instb in qc_before.data:
                    if (
                        instb[0].name == "measure"
                        and instb[2][0].index == inst[2][0].index
                    ):
                        logical_q = instb[1][0].index
                        break
            logical_physical_mapping[logical_q] = inst[1][0].index
    return logical_physical_mapping


def _add_measure_all_if_none(qc: Any) -> Any:
    for inst in qc.data:
        if inst[0].name == "measure":
            break
    else:
        qc.measure_all()
    return qc


def qiskit_compile(
    circuit: Any,
    info: Optional[Dict[str, Any]] = None,
    output: str = "tc",
    compiled_options: Optional[Dict[str, Any]] = None,
) -> Any:
    """
    compile the circuit using ``qiskit.transpile`` method with some tricks and hacks

    :param circuit: circuit in ``tc.Circuit`` or ``qiskit.QuantumCircuit`` form
    :type circuit: Any
    :param info: info for qubit mappings, defaults to None
    :type info: Optional[Dict[str, Any]], optional
    :param output: output circuit format, defaults to "tc"
    :type output: str, optional
    :param compiled_options: ``qiskit.transpile`` options in a dict, defaults to None
    :type compiled_options: Optional[Dict[str, Any]], optional
    :return: Tuple containing the output circuit and the qubit mapping info dict
    :rtype: Any
    """
    from qiskit.compiler import transpile
    from qiskit.transpiler.passes import RemoveBarriers

    if isinstance(circuit, AbstractCircuit):
        circuit = circuit.to_qiskit(enable_instruction=True)
    elif isinstance(circuit, str):
        circuit = qiskit_from_qasm_str_ordered_measure(circuit)
    # else qiskit circuit
    circuit = _add_measure_all_if_none(circuit)
    if compiled_options is None:
        compiled_options = {
            "basis_gates": ["h", "rz", "cx"],
            "optimization_level": 2,  # 3 can induce bugs...
        }
    ncircuit = transpile(circuit, **compiled_options)
    ncircuit = RemoveBarriers()(ncircuit)

    if output.lower() in ["qasm", "openqasm"]:
        r0 = ncircuit.qasm()

    elif output.lower() in ["qiskit", "ibm"]:
        r0 = ncircuit

    elif output.lower() in ["tc", "tensorcircuit"]:
        s = _free_pi(ncircuit.qasm())
        r0 = Circuit.from_openqasm(
            s,
            keep_measure_order=True,
        )

    else:
        raise ValueError("Unknown output format: %s" % output)

    r1 = {}
    nlpm = _get_logical_physical_mapping_from_qiskit(ncircuit, circuit)
    # new_logical_physical_mapping
    if info is not None and "logical_physical_mapping" in info:
        r1["logical_physical_mapping"] = {
            k: nlpm[v] for k, v in info["logical_physical_mapping"].items()
        }
    else:
        r1["logical_physical_mapping"] = nlpm
    if info is not None and "positional_logical_mapping" in info:
        r1["positional_logical_mapping"] = info["positional_logical_mapping"]
    else:  # info is none, assume circuit is the logical circuit
        r1["positional_logical_mapping"] = _get_positional_logical_mapping_from_qiskit(
            circuit
        )
    # TODO(@refraction-ray): more info to be added into r1 dict

    return (r0, r1)
